import re
import requests
from .config import scrape_value_by_key, print_msg
from .prometheus import get_ignored_words


def grafana_http_request(path, api_key, base_url="http://localhost:3000"):
    if not api_key:
        print_msg(msg_content="Missing API Key",
                  msg_type='error', terminate=True)
    url = f"{base_url}{path}"
    response = requests.get(url, headers={
        "Authorization": f"Bearer {api_key}"
    })
    if 200 <= response.status_code < 400:
        print_msg(
            msg_content=f"Successful response from {url}", msg_type='log')
        return response
    if response.status_code == 401:
        print_msg(f"API Key is invalid", response.text,
                  'error')
    else:
        print_msg(f"Unknown response", response.text, 'error')
    return response


def get_metrics_from_expr(expression, ignored_words_list):
    # Remove all [] and {}
    regex_patterns = [
        r"\{(.+?)\}",
        r"\[(.+?)\]",
    ]
    for pattern in regex_patterns:
        results = re.findall(pattern, expression)
        if results:
            results = list(set(results))
            results = [
                result.strip() for result in results
                if result and result != " "
            ]
            for result in results:
                expression = expression.replace(result, " ")

    # Remove all the ignored_words
    for item in ignored_words_list:
        pattern = r"[\(\)\{\}]" + item + r"[ \(\)\{\}]"
        results = re.findall(pattern, expression)
        if results:
            results = list(set(results))
            results = [
                result.strip() for result in results
                if result and result != " "
            ]
            for result in results:
                expression = expression.replace(result, " ")

    # Remove leftovers
    # TODO: Avoid this loop, cleanup in previous step
    math_signs = ["*", "-", "+", "^", "/", "]",
                  "[", "{", "}", "__", ",", "=", "\\", "'", '"', "_over_time"]
    for sign in math_signs:
        expression = expression.replace(sign, " ")
    metrics = expression.split()
    return metrics


def get_metrics_list(base_url, api_key):
    ignored_words = get_ignored_words()
    if len(ignored_words):
        print_msg(
            msg_content=f"Found {len(ignored_words)} words to ignore in expressions")
    else:
        print_msg(msg_content="No words to ignore, that's weird",
                  msg_error="warning")
    dashboards = grafana_http_request(
        "/api/search?query=",
        api_key,
        base_url
    ).json()
    data = {
        "dashboards": {}
    }
    for dashboard in dashboards:
        dashboard_body = grafana_http_request(
            f"/api/dashboards/uid/{dashboard['uid']}",
            api_key,
            base_url
        ).json()
        dashboard_gnetid = dashboard_body['dashboard']['gnetId'] if 'gnetId' in dashboard_body[
            'meta'] and dashboard_body['meta']['gnetId'] else "null"
        dashboard_name = dashboard_body['meta']['slug'] if 'slug' in dashboard_body[
            'meta'] and dashboard_body['meta']['slug'] else "null"
        print_msg(msg_content=f"Getting metrics from {dashboard_name}")
        expressions = scrape_value_by_key(dashboard_body, "expr", str)
        expressions += scrape_value_by_key(dashboard_body, "query", str)
        dashboard_metrics = []
        for expression in expressions:
            try:
                expr_metrics = get_metrics_from_expr(expression, ignored_words)
            except TypeError:
                print_msg(msg_content=f"The following expression is corrupted",
                          data=expression, msg_type='e')
            if expr_metrics:
                for metric in expr_metrics:
                    try:
                        float(metric)
                    except:  # noqa: 722
                        if len(metric) > 5 and "/" not in metric:
                            dashboard_metrics.append(metric)
        dashboard_metrics = sorted(list(set(dashboard_metrics)))
        data['dashboards'][dashboard_name] = dict()
        data['dashboards'][dashboard_name]['metrics'] = dashboard_metrics
        data['dashboards'][dashboard_name]['gnet_id'] = dashboard_gnetid
        data['dashboards'][dashboard_name]['num_metrics'] = len(
            dashboard_metrics)
        print_msg(
            msg_content=f"Found {data['dashboards'][dashboard_name]['num_metrics']} metrics")

    all_metrics = []
    for dashboard_name in data['dashboards']:
        all_metrics += data['dashboards'][dashboard_name]['metrics']
    data['all_metrics'] = sorted(list(set(all_metrics)))
    data['all_metrics_num'] = len(data['all_metrics'])
    print_msg(
        msg_content=f"Found a total of {data['all_metrics_num']} unique metrics to keep")
    return data



import json
import click
from .grafana import get_metrics_list
from .config import print_msg, is_docker, pass_config
from .prometheus import apply_yaml


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        app_aliases = {
            "m": "main",
            "p": "prometheus",
            "g": "grafana",
        }
        action_aliases = {
            "a": "apply",
            "g": "get",
            "d": "delete",
            "l": "list",
        }
        if len(cmd_name) == 2:
            words = []
            if cmd_name[0] in app_aliases:
                words.append(app_aliases[cmd_name[0]])
            if cmd_name[1] in action_aliases:
                words.append(action_aliases[cmd_name[1]])
            if len(words) == 2:
                cmd_name = "-".join(words)

        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")


# @click.group()
@click.command(cls=AliasedGroup)
@pass_config
@click.option('--ci', '-ci', is_flag=True, help="Use this flag to avoid confirmation prompts")  # noqa: E501
def cli(config, ci):
    """No confirmation prompts"""  # noqa: E501
    if is_docker():
        ci = True
    config.ci = ci  # noqa: F821


@cli.command()
@click.option('--grafana-url', '-gurl', default='http://localhost:3000', prompt=True, required=False, show_default=True, type=str)
@click.option('--grafana-api-key', '-gkey', prompt=True, required=True, hide_input=True, type=str)
@click.option('--output-file-path', '-o', default='./.metrics.json', show_default=True, required=False, type=str)
# @click.option('--file', '-f', prompt=False, default=".", help="Output metrics.json file to pwd")
def grafana_list(grafana_url, grafana_api_key, output_file_path):
    """Provide Grafana UrL and Grafana API Key (Viewer)\n
Returns a list of metrics that are used in all dashboards"""
    if "http" not in grafana_url:
        print_msg(
            msg_content="Grafana URL must contain 'http' or 'https'", msg_type="error")
    metrics = get_metrics_list(grafana_url, grafana_api_key)
    with open(output_file_path, 'w') as file:
        json.dump(metrics, file, indent=2, sort_keys=True)


@cli.command()
@click.option('--prom-yaml-path', '-ppath', default='docker-compose/prometheus.yml', prompt=True, required=True, show_default=False, type=str)
@click.option('--metrics-json-path', '-mjpath', default='./.metrics.json', show_default=True, prompt=True, required=False, type=str)
@click.option('--create-backup-file', '-b', is_flag=True, default=True, required=False)
def prometheus_apply(prom_yaml_path, metrics_json_path, create_backup_file):
    apply_yaml(prom_yaml_path, metrics_json_path, create_backup_file)



import json
import requests
from bs4 import BeautifulSoup
import yaml
from .config import print_json, print_msg


def request_words(url, selector, replace_str="()", replace_with=""):
    response = requests.get(url)
    if 200 <= response.status_code < 400:
        print_msg(
            msg_content=f"Successfully got words from {url}", msg_type='log')
    else:
        print_msg(
            msg_content=f"Failed to reach {url}, have you provided the '--web.enable-admin-api' in Prometheus?")
    html_page_text = response.text
    soup = BeautifulSoup(html_page_text, 'html.parser')
    return [
        item.string.replace(replace_str, replace_with)
        for item in soup.select(selector)
    ]


def get_ignored_words():
    """Get the list of words to ignore when scraping from Grafana"""
    print_msg(
        msg_content="Getting the list of words to ignore when scraping from Grafana")
    words_list = []

    prometheus_urls = [
        {
            "url": "https://prometheus.io/docs/prometheus/latest/querying/functions/",
            "selectors": ".toc-right ul li code"
        },
        {
            "url": "https://prometheus.io/docs/prometheus/latest/querying/operators/",
            "selectors": "ul li code"
        }
    ]
    for item in prometheus_urls:
        words_list += request_words(item["url"], item["selectors"])
    return sorted(list(set(words_list)))


def apply_yaml(prom_yaml_path, metrics_json_path, create_backup_file=True):
    print_msg(msg_content=f"Reading documents from {prom_yaml_path}")
    with open(prom_yaml_path, "r") as file:
        prom_yaml = file.read()
    prom_documents = yaml.safe_load_all(prom_yaml)
    prom_doc = None
    frigga_doc = None
    for doc in prom_documents:
        if 'scrape_configs' in doc:
            prom_doc = doc
        elif 'name' in doc and doc['name'] == 'frigga':
            frigga_doc = doc
    if not prom_doc:
        print_msg(
            msg_content="Missing 'scrape_configs' in prometheus.yml", msg_type="error")
    if not frigga_doc:
        print_msg(
            msg_content="Missing frigga document in prometheus.yml", msg_type="error")
    print_msg(msg_content=f"Found relevant documents")
    print_msg(msg_content=f"Loading metrics.json")
    with open(metrics_json_path, "r") as file:
        metrics_dict = json.load(file)

    print_msg(msg_content=f"Generating metrics_relabel_configs in memory")
    metric_relabel_configs = []
    for metric in metrics_dict['all_metrics']:
        metric_relabel_configs.append({
            "source_labels": ['__name__'],
            "regex": f"^{metric}",
            "target_label": "__tmp_keep_me",
            "replacement": True
        })
    metric_relabel_configs.append({
        "source_labels": ["__tmp_keep_me"],
        "regex": True,
        "action": "keep"
    })
    for job in prom_doc['scrape_configs']:
        if job['job_name'] not in frigga_doc['exclude_jobs']:
            job['metric_relabel_configs'] = metric_relabel_configs

    noalias_dumper = yaml.dumper.SafeDumper
    noalias_dumper.ignore_aliases = lambda self, data: True

    # backup old prom yaml
    if create_backup_file:
        print_msg(msg_content=f"Creating a backup file for {prom_yaml_path}")
        with open(prom_yaml_path, 'r') as source_fs:
            with open(f"{prom_yaml_path}.bak.yml", 'w') as target_fs:
                target_fs.write(source_fs.read())

    # write new prom frigga yaml
    print_msg(
        msg_content=f"Writing the new metrics_relabel_configs to {prom_yaml_path}")
    with open(prom_yaml_path, 'w') as fs:
        data = yaml.dump_all(
            documents=[prom_doc, frigga_doc],
            stream=fs, default_flow_style=False,
            Dumper=noalias_dumper,
            indent=2
        )
    print_msg(
        msg_content=f"Done! Now reload {prom_yaml_path} with 'docker exec $PROM_CONTAINER_NAME kill -HUP 1'")



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    readme = fh.read()

with open("version", "r") as fh:
    version = fh.read()

setup(
    author='Meir Gabay',
    author_email='unfor19@gmail.com',
    name='frigga',
    license='MIT',
    long_description_content_type='text/markdown',
    description='Scrape only relevant metrics in Prometheus, according to your Grafana dashboards',  # noqa: E501
    long_description=readme,
    version=f'{version}',
    packages=find_packages(),
    keywords='prometheus grafana dashboard metrics monitor cloud',
    include_package_data=True,
    url='https://github.com/unfor19/frigga',
    download_url=f'https://github.com/unfor19/frigga/archive/v{version}.tar.gz',  # noqa: E501
    install_requires=[
        'Click>=7.1.1',
        'beautifulsoup4>=4.9.1',
        'requests>=2.23.0',
        'PyYAML>=5.3.1'
    ],
    entry_points='''
        [console_scripts]
        frigga=scripts.frigga:cli
    ''',
    setup_requires=[
        'setuptools>=44.1.0',
        'wheel>=0.34.2',
        'twine==3.1.1',
        'docutils>=0.16'
    ],
    classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8'
  ],
)



import json
import os
import click


def is_docker():
    path = '/proc/self/cgroup'
    return (
        os.path.exists('/.dockerenv') or
        os.path.isfile(path) and any('docker' in line for line in open(path))
    )


def print_msg(msg_content, data=None, msg_type='l', terminate=False):
    if msg_type == "data" or msg_type == "d":
        msg = f">> {msg_content}\nData:\n"
    elif msg_type == "error" or msg_type == "e":
        msg = f">> [ERROR] {msg_content}"
        terminate = True
    elif msg_type == "warning" or msg_type == "w":
        msg = f">> [WARNING] {msg_content}"
    elif msg_type == "log" or msg_type == "l":
        msg = f">> [LOG] {msg_content}"

    if data:
        if isinstance(data, dict) or isinstance(data, list):
            print(f"{msg}\n{json.dumps(data, indent=2, sort_keys=True)}")
        else:
            print(f"{msg} {data}")
    else:
        print(msg)

    if terminate:
        exit()


def print_json(obj):
    print(json.dumps(obj, indent=2, sort_keys=True))


def scrape_value_by_key(obj, search_key, type_key, my_list=[]):
    """Provide and obj and key to search
Returns a list of values, according to given key"""
    if isinstance(obj, list) and len(obj):
        for item in obj:
            scrape_value_by_key(item, search_key, type_key, my_list)
    elif isinstance(obj, dict):
        if search_key in obj \
                and obj[search_key] \
                and isinstance(obj[search_key], type_key):
            return my_list.append(obj[search_key])
        else:
            for key, value in obj.items():
                if (isinstance(value, list) or isinstance(value, dict)):
                    scrape_value_by_key(value, search_key, type_key, my_list)
    return my_list


class Config(object):
    def __init__(self, raise_error=True):
        self.verbose = False
        self.errors = 0
        self.errors_msg = ""
        self.ci = False


pass_config = click.make_pass_decorator(Config, ensure=True)
